// const { Middleware } = require('swagger-express-middleware')
const http = require('http')
const fs = require('fs')
const path = require('path')
const swaggerUI = require('swagger-ui-express')
const jsYaml = require('js-yaml')
const express = require('express')
const cors = require('cors')
const cookieParser = require('cookie-parser')
const bodyParser = require('body-parser')
const OpenApiValidator = require('express-openapi-validator')
const logger = require('./logger')
const config = require('./config')

const serviceManager = require('./services/Service.js')

const { getBasePath } = require('./utils/swaggerUtils')
const { updateAPISpecification } = require('./utils/fixes')

const { entrypoint } = require('./utils/entrypoint')

const cleanPath = (path) => path.replace(/\/\//,'/')

class ExpressServer {
  constructor(port, openApiYaml) {
    this.port = port;
    this.app = express();
    this.openApiPath = openApiYaml;
    try {
    
      this.schema = jsYaml.safeLoad(fs.readFileSync(openApiYaml))
      
      updateAPISpecification(this.schema, openApiYaml)

    } catch (e) {
      logger.error('failed to start Express Server', e.message)
    }

    this.setupMiddleware()
  
  }

  setupMiddleware() {

    const basePath = getBasePath().replace(/\/$/,'')

    // this.setupAllowedMedia();
    this.app.use(cors())
    // this.app.use(bodyParser.json({ limit: '14MB' }))
    this.app.use(express.json())
    this.app.use(express.text())
    this.app.use(express.urlencoded({ extended: false }))
    this.app.use(cookieParser())
    
    //Simple test to see that the server is up and responding
    this.app.get('/hello', (req, res) => res.send(`Hello World. path: ${this.openApiPath}`))
    
    const openapi=config.OPENAPI || '/openapi'

    //Send the openapi document *AS GENERATED BY THE GENERATOR*
    this.app.get(cleanPath(`${basePath}${openapi}`), (req, res) => res.sendFile((path.join(__dirname, 'api', 'openapi.yaml'))))

    //View the openapi document in a visual interface. Should be able to test from this page
    this.app.use(cleanPath(`${basePath}/api-docs`), swaggerUI.serve, swaggerUI.setup(this.schema))

    this.app.get('/login-redirect', (req, res) => {
      res.status(200)
      res.json(req.query)
    })

    this.app.get('/oauth2-redirect.html', (req, res) => {
      res.status(200)
      res.json(req.query)
    })
    
    this.app.get(`${openapi}`, function(req, res) {
      res.redirect(cleanPath(`${basePath}/${openapi}`))
    })

    this.app.get('/api-docs', function(req, res) {
      res.redirect(cleanPath(`${basePath}/api-docs`))
    })

    this.app.use(cleanPath(`${basePath}/entrypoint`), entrypoint)

    this.app.get('/', function(req, res) {
      res.redirect(cleanPath(`${basePath}/entrypoint`))
    })

    this.app.get(cleanPath(`${basePath}/`), function(req, res) {
      res.redirect(cleanPath(`${basePath}/entrypoint`))
    })

    this.app.get(cleanPath(`${basePath}/*/${config.MONITOR_PATH}/:id`), async function(req, res) {
      let context = {
        classname:    config.MONITOR,
        operationId: 'internalRetrieve',
        method:       "get",
        request:     { url: req.url }
      }

      let args = {
        id: req.params.id
      }

      try {
        let result = await serviceManager.show(args, context)
        res.json(result.payload, null, 4)
      } catch(e) {
        res.status(e.code).json({ e })
      }
    })

    const apiTimeout = 5 * 1000
    if(false) this.app.use((req, res, next) => {
        // Set the timeout for all responses
        res.setTimeout(apiTimeout, () => {
            let err = new Error('Service Unavailable')
            err.status = 503
            next(err)
        })
    })

  }

  launch() {
    
    try {
      
      this.app.use( OpenApiValidator.middleware({
            apiSpec: this.openApiPath,
            operationHandlers: path.join(__dirname),
            validateRequests: true, // (default)
            validateResponses: false, // false by default
            unknownFormats: ['base64']
        }))
        
      // this.app.use(validator)

      logger.info(`validator installed`)
  
      this.app.use((err, req, res, next) => {
        // format errors
        res.status(err.status || 500).json({
          message: err.message || err,
          errors: err.errors || '',
        })
      })

      const server = http.createServer(this.app)
      server.listen(this.port)

      logger.info(`Listening on port ${this.port}`)
       
    } catch(e) {
        logger.info(`launch error: ` + e)
    }

  }


  async close() {
    if (this.server !== undefined) {
      await this.server.close();
      logger.info(`Server on port ${this.port} shut down`)
    }
  }
}

module.exports = ExpressServer;
